"""
This modules contains the logic to parse and validate JSON files in the OpenXC
message set format.
"""

from __future__ import print_function
from collections import defaultdict
import operator
import logging

from .xml_to_json import merge_database_into_mapping, parse_database
from .structures import Command, CanBus, DiagnosticMessage
from openxc.utils import fatal_error, merge, find_file, \
        load_json_from_search_path

LOG = logging.getLogger(__name__)


class MessageSet(object):
    def __init__(self, name):
        self.name = name
        self.buses = defaultdict(CanBus)
        self.initializers = []
        self.loopers = []
        self.commands = []
        self.extra_sources = set()

    def valid_buses(self):
        valid_buses = [bus for bus in self.buses.values() if bus.valid()]
        return sorted(valid_buses, key=operator.attrgetter('controller'))

    def active_diagnostic_messages(self):
        for message in self.all_diagnostic_messages():
            if message.enabled:
                yield message

    def all_diagnostic_messages(self):
        for message in sorted(self.diagnostic_messages,
                key=lambda x: (x.id, x.mode, x.pid)):
            yield message

    def active_messages(self):
        for bus in self.buses.values():
            for message in bus.active_messages():
                yield message

    def all_messages(self):
        for bus in self.buses.values():
            for message in bus.sorted_messages():
                yield message

    def enabled_signals(self):
        for signal in self.all_signals():
            if signal.enabled:
                yield signal

    def active_commands(self):
        for command in self.all_commands():
            if command.enabled:
                yield command

    def all_commands(self):
        for command in sorted(self.commands, key=operator.attrgetter('name')):
            yield command

    def all_signals(self):
        for message in self.all_messages():
            for signal in message.sorted_signals():
                yield signal

    def validate(self):
        return self.validate_name() and self.validate_messages()

    def validate_messages(self):
        valid = True
        for message in self.active_messages():
            valid = valid and message.validate()
        for signal in self.enabled_signals():
            valid = valid and signal.validate()
        return valid

    def validate_name(self):
        if self.name is None:
            LOG.warning("missing message set (%s)" % self.name)
            return False
        return True

    def lookup_message_index(self, message):
        for i, candidate in enumerate(self.active_messages()):
            if candidate.id == message.id:
                return i

    def lookup_bus(self, name=None, controller=None):
        if name is not None:
            return self.buses.get(name, None)
        else:
            for bus in self.buses.values():
                if bus.controller == controller:
                    return bus

    def lookup_bus_index(self, bus_name):
        for index, bus in enumerate(self.valid_buses()):
            if bus_name == bus.name:
                return index
        return None

    def _message_count(self):
        return len(list(self.all_messages()))


class JsonMessageSet(MessageSet):
    @classmethod
    def parse(cls, filename, search_paths=None, skip_disabled_mappings=False):
        search_paths = search_paths or []

        data = load_json_from_search_path(filename, search_paths)

        # Flatten all parent files into one dict
        while len(data.get('parents', [])) > 0:
            for parent_filename in data.get('parents', []):
                parent_data = load_json_from_search_path(parent_filename,
                        search_paths)
                # Merge data *into* parents, so we keep any overrides
                data = merge(parent_data, data)
                data['parents'].remove(parent_filename)

        message_set = cls(data.get('name', 'generic'))

        message_set.initializers = data.get('initializers', [])
        message_set.loopers = data.get('loopers', [])
        message_set.buses = cls._parse_buses(data)
        message_set.bit_numbering_inverted = data.get(
                'bit_numbering_inverted', None)
        message_set.extra_sources = set(data.get('extra_sources', set()))

        mapping_config = message_set._parse_mappings(data, search_paths,
                skip_disabled_mappings)
        message_set.initializers.extend(mapping_config['initializers'])
        message_set.loopers.extend(mapping_config['loopers'])
        message_set.extra_sources.update(mapping_config['extra_sources'])

        # TODO could we parse the top level config file as if it were a mapping
        # and avoid duplicating this code?
        for message_id, message in data.get('messages', {}).items():
            message['id'] = message_id
            mapping_config['messages'].append(message)

        mapping_config['diagnostic_messages'].extend(data.get('diagnostic_messages', []))
        message_set.diagnostic_messages = cls._parse_diagnostic_messages(
                message_set, mapping_config['diagnostic_messages'])

        mapping_config['commands'].extend(data.get('commands', []))
        message_set.commands = cls._parse_commands(mapping_config['commands'])
        message_set._parse_messages(mapping_config['messages'])

        return message_set

    @classmethod
    def _parse_commands(cls, commands):
        return [Command(**command_data) for command_data in commands]

    @classmethod
    def _parse_diagnostic_messages(cls, message_set, messages):
        return [DiagnosticMessage(message_set, **message_data)
                for message_data in messages]

    @classmethod
    def _parse_buses(cls, data):
        buses = {}
        for bus_name, bus_data in data.get('buses', {}).items():
            buses[bus_name] = CanBus(name=bus_name,
                    default_raw_can_mode=data.get('raw_can_mode', "off"),
                    default_max_message_frequency=data.get('max_message_frequency', 0),
                    **bus_data)
            if buses[bus_name].speed is None:
                fatal_error("Bus %s is missing the 'speed' attribute" %
                        bus_name)
        return buses

    @classmethod
    def _parse_database(self, database_filename):
        if getattr(self, '_database_cache', None) is None:
                self._database_cache = {}
        if self._database_cache.get(database_filename, None) is None:
            self._database_cache[database_filename] = parse_database(
                    database_filename)
        return database_filename, self._database_cache[database_filename]

    def _parse_mappings(self, data, search_paths, skip_disabled_mappings):
        all_messages = []
        all_diagnostic_messages = []
        all_commands = []
        all_extra_sources = set()
        all_initializers = []
        all_loopers = []

        for mapping in data.get('mappings', []):
            if 'mapping' not in mapping:
                fatal_error("Mapping is missing the mapping file path")

            mapping_enabled = mapping.get('enabled', True)
            if not mapping_enabled:
                LOG.warning("Mapping '%s' is disabled" % mapping['mapping'])
                if skip_disabled_mappings:
                    continue
            LOG.warning("Adding mapping '%s'" % mapping['mapping'])

            bus_name = mapping.get('bus', None)
            if bus_name is None:
                LOG.warning("No default bus associated with '%s' mapping" %
                        mapping['mapping'])
            elif bus_name not in self.buses:
                fatal_error("Bus '%s' (from mapping %s) is not defined" %
                        (bus_name, mapping['mapping']))
            elif not self.buses[bus_name].valid():
                LOG.warning("Mapping '%s' is disabled because bus '%s' " % (
                    mapping['mapping'], bus_name) +
                        "is not associated with a CAN controller")
                mapping['enabled'] = False
                if skip_disabled_mappings:
                    continue

            mapping_data = load_json_from_search_path(mapping['mapping'],
                    search_paths)

            commands = mapping_data.get('commands', [])
            for command in commands:
                command.setdefault('enabled', mapping_enabled)
            all_commands.extend(commands)

            diagnostic_messages = mapping_data.get('diagnostic_messages', [])
            for message in diagnostic_messages:
                message.setdefault('bus', bus_name)
                message.setdefault('enabled', mapping_enabled)
            all_diagnostic_messages.extend(diagnostic_messages)

            if mapping_enabled:
                all_initializers.extend(mapping_data.get('initializers', []))
                all_loopers.extend(mapping_data.get('loopers', []))
                all_extra_sources.update(
                        set(mapping_data.get('extra_sources', set())))

            messages = mapping_data.get('messages', {})
            if len(messages) == 0:
                LOG.warning("Mapping file '%s' doesn't define any messages"
                        % mapping['mapping'])

            if 'database' in mapping:
                database_filename, database_tree = self._parse_database(
                                find_file(mapping['database'], search_paths))
                messages = merge(merge_database_into_mapping(database_filename,
                    database_tree, messages).get('messages', {}), messages)
                if mapping.get('bit_numbering_inverted', None) is None:
                    LOG.warning("The bit number inversion setting is undefined "
                            "for the mapping '%s', but it " % mapping['mapping'] +
                            "is database-backed - assuming inverted")
                    for message in messages.values():
                        message.setdefault('bit_numbering_inverted',
                                True)

            for message_id, message in messages.items():
                message['id'] = message_id
                message.setdefault('bus', bus_name)
                message.setdefault('enabled', mapping_enabled)
            all_messages.extend(messages.values())

        return {'messages': all_messages,
                'commands': all_commands,
                'diagnostic_messages': all_diagnostic_messages,
                'initializers': all_initializers,
                'extra_sources': all_extra_sources,
                'loopers': all_loopers}

    def _parse_messages(self, messages, default_bus=None):
        for message_data in messages:
            message_data.setdefault('bit_numbering_inverted',
                    self.bit_numbering_inverted)
            message = self.buses[message_data['bus']].get_or_create_message(
                    message_data['id'])
            message.message_set = self
            message.merge_message(message_data)



"""Controller implementation for a virtual serial device."""
from __future__ import absolute_import
import time

from .base import Controller


class SerialControllerMixin(Controller):
    """An implementation of a Controller type that connects to a virtual serial
    device.

    This class acts as a mixin, and expects ``self.device`` to be an instance
    of ``serial.Serial``.

    TODO Bah, this is kind of weird. refactor the relationship between
    sources/controllers.
    """

    WAITIED_FOR_CONNECTION = False

    def write_bytes(self, data):
        return self.device.write(data)

    def complex_request(self, request, blocking=True):
        if not self.WAITIED_FOR_CONNECTION:
            # TODO need to wait until device is connected or errors out
            # that may be a bug in the bluetooth stack, see
            # https://bugzilla.redhat.com/show_bug.cgi?id=1060457
            time.sleep(5)
        return super(SerialControllerMixin, self).complex_request(request, blocking)



from .base import DataSource, DataSourceError
from .usb import UsbDataSource
from .serial import SerialDataSource
from .trace import TraceDataSource
from .network import NetworkDataSource
from .bluetooth import BluetoothVehicleInterface



"""Contains the abstract interface for sending commands back to a vehicle
interface.
"""
import numbers
import time
import threading
import binascii

try:
    from Queue import Queue
    from Queue import Empty
except ImportError:
    # Python 3
    from queue import Queue
    from queue import Empty

class ResponseReceiver(object):
    """All commands to a vehicle interface are asynchronous. This class is used to
    wait for the response for a particular request in a thread. Before making a
    request, a ResponseReceiver is created to wait for the response. All
    responses received from the VI (which may or may not be in response to this
    particular command) are passed to the ResponseReceiver, until it either
    times out waiting or finds a matching response.

    The synchronization mechanism is a multiprocessing Queue. The
    ResponseReceiver blocks waiting on a new response to be added to the queue,
    and the vehicle interface class puts newly received responses in the queues
    of ResponseReceivers as they arrive.
    """

    COMMAND_RESPONSE_TIMEOUT_S = .5

    def __init__(self, queue, request, quit_after_first=True):
        """Construct a new ResponseReceiver.

        queue - A multithreading queue that this receiver will pull potential responses from.
        request - The request we are trying to match up with a response.
        """
        self.request = request
        self.queue = queue
        self.responses = []
        self.running = True
        self.quit_after_first = quit_after_first

    def _response_matches_request(self, response):
        """Inspect the given response and return true if it's a response to this
        ResponseReceiver's request.

        This implementation is the base class no-op - it returns True for any
        response. You probably want to override this in a subclass.

        response - the response to inspect.
        """
        return True

    def wait_for_responses(self):
        """Block the thread and wait for the response to the given request to
        arrive from the VI. If no matching response is received in
        COMMAND_RESPONSE_TIMEOUT_S seconds, returns anyway.
        """

        self.thread.join(self.COMMAND_RESPONSE_TIMEOUT_S)
        self.running = False

        return self.responses

    def start(self):
        self.thread = threading.Thread(target=self.handle_responses)
        self.thread.start()

    def handle_responses(self):
        """Block and wait for responses to this object's original request, or
        until a timeout (self.COMMAND_RESPONSE_TIMEOUT_S).

        This function is handy to use as the target function for a thread.

        The responses received (or None if none was received before the timeout)
        is stored in a list at self.responses.
        """
        while self.running:
            try:
                response = self.queue.get(
                        timeout=self.COMMAND_RESPONSE_TIMEOUT_S)
                if self._response_matches_request(response):
                    self.responses.append(response)
                    if self.quit_after_first:
                        self.running = False
                self.queue.task_done()
            except Empty:
                break

class CommandResponseReceiver(ResponseReceiver):
    """A receiver that matches the 'command' field in responses to the
    original request.
    """

    def _response_matches_request(self, response):
        """Return true if the 'command' field in the response matches the
        original request.
        """
        return response.get('command_response', None) == self.request['command']

class DiagnosticResponseReceiver(ResponseReceiver):
    """A receiver that matches the bus, ID, mode and PID from a
    diagnostic request to an incoming response.
    """

    def __init__(self, queue, request):
        super(DiagnosticResponseReceiver, self).__init__(queue, request,
                quit_after_first=False)
        # Make sure to key off of the diagnostic request, not the command to
        # create the request
        self.diagnostic_request = request['request']

    def _response_matches_request(self, response):
        """Return true if the response is to a diagnostic request, and the bus,
        id, mode match. If the request was successful, the PID echo is also
        checked.
        """
        # Accept success/failure command responses
        if super(DiagnosticResponseReceiver,
                self)._response_matches_request(response):
            return True

        if ('bus' in self.diagnostic_request and
                response.get('bus', None) != self.diagnostic_request['bus']):
            return False
        if (self.diagnostic_request['id'] != 0x7df and
                response.get('id', None) != self.diagnostic_request['id']):
            return False

        if (response.get('success', True) and
                response.get('pid', None) !=
                    self.diagnostic_request.get('pid', None)):
            return False

        return response.get('mode', None) == self.diagnostic_request['mode']


class Controller(object):
    """A Controller is a physical vehicle interface that accepts commands to be
    send back to the vehicle. This class is abstract, and implementations of the
    interface must define at least the ``write_bytes`` method.
    """

    def _prepare_response_receiver(self, request,
            receiver_class=CommandResponseReceiver):
        queue = Queue()

        self.open_requests = getattr(self, 'open_requests', [])
        self.open_requests.append(queue)

        receiver = receiver_class(queue, request)
        receiver.start()
        # Give it a brief moment to get started so we make sure get the response
        time.sleep(.2)
        return receiver

    def complex_request(self, request, wait_for_first_response=True):
        """Send a compound command request to the interface over the normal data
        channel.

        request - A dict storing the request to send to the VI. It will be
            serialized to the currently selected output format.
        wait_for_first_response - If true, this function will block waiting for
            a response from the VI and return it to the caller. Otherwise, it
            will send the command and return immediately and any response will
            be lost.
        """
        receiver = self._prepare_response_receiver(request,
                receiver_class=CommandResponseReceiver)
        self._send_complex_request(request)

        responses = []
        if wait_for_first_response:
            responses = receiver.wait_for_responses()
        return responses

    def _send_complex_request(self, request):
        self.write_bytes(self.streamer.serialize_for_stream(request))

    @classmethod
    def _build_diagnostic_request(cls, message_id, mode, bus=None, pid=None,
            frequency=None, payload=None, decoded_type=None):
        request = {
            'command': "diagnostic_request",
            'request': {
                'id': message_id,
                'mode': mode
            }
        }

        if bus is not None:
            request['request']['bus'] = bus
            request['request']['mode'] = mode
        if payload is not None and len(payload) > 0:
            # payload must be a bytearray
            request['request']['payload'] = "0x%s" % binascii.hexlify(payload)
        if pid is not None:
            request['request']['pid'] = pid
        if frequency is not None:
            request['request']['frequency'] = frequency
        if decoded_type is not None:
            request['request']['decoded_type'] = decoded_type

        return request

    def delete_diagnostic_request(self, message_id, mode, bus=None, pid=None):
        request = self._build_diagnostic_request(message_id, mode, bus, pid)
        request['action'] = 'cancel'
        return self._check_command_response_status(request)

    def create_diagnostic_request(self, message_id, mode, bus=None, pid=None,
            frequency=None, payload=None, wait_for_ack=True,
            wait_for_first_response=False, decoded_type=None):
        """Send a new diagnostic message request to the VI

        Required:

        message_id - The message ID (arbitration ID) for the request.
        mode - the diagnostic mode (or service).

        Optional:

        bus - The address of the CAN bus controller to send the request, either
            1 or 2 for current VI hardware.
        pid - The parameter ID, or PID, for the request (e.g. for a mode 1
            request).
        frequency - The frequency in hertz to add this as a recurring diagnostic
            requests. Must be greater than 0, or None if it is a one-time
            request.
        payload - A bytearray to send as the request's optional payload. Only
            single frame diagnostic requests are supported by the VI firmware in
            the current version, so the payload has a maximum length of 6.
        wait_for_ack - If True, will wait for an ACK of the command message.
        wait_for_first_response - If True, this function will block waiting for
            a diagnostic response to be received for the request. It will return
            either after timing out or after 1 matching response is received -
            there may be more responses to functional broadcast requests that
            arrive after returning.

        Returns a tuple of
            ([list of ACK responses to create request],
                [list of diagnostic responses received])

        """

        request = self._build_diagnostic_request(message_id, mode, bus, pid,
                frequency, payload, decoded_type)

        diag_response_receiver = None
        if wait_for_first_response:
            diag_response_receiver = self._prepare_response_receiver(
                    request, DiagnosticResponseReceiver)

        request['action'] = 'add'
        ack_responses = self.complex_request(request, wait_for_ack)

        diag_responses = None
        if diag_response_receiver is not None:
            diag_responses = diag_response_receiver.wait_for_responses()

        return ack_responses, diag_responses

    def _check_command_response_status(self, request):
        responses = self.complex_request(request)
        return len(responses) > 0 and responses[0]['status']

    def set_passthrough(self, bus, enabled):
        """Control the status of CAN message passthrough for a bus.

        Returns True if the command was successful.
        """
        request = {
            "command": "passthrough",
            "bus": bus,
            "enabled": enabled
        }
        return self._check_command_response_status(request)

    def set_payload_format(self, payload_format):
        """Set the payload format for messages sent to and from the VI.

        Returns True if the command was successful.
        """
        request = {
            "command": "payload_format",
            "format": payload_format
        }
        status = self._check_command_response_status(request)
        # Always change the format regardless because if it was already in the
        # right format, the command will have failed.
        self.format = payload_format
        return status

		
    def rtc_configuration(self, unix_time):
        """Set the Unix time if RTC is supported on the device.

        Returns True if the command was successful.
        """
        request = {
            "command": "rtc_configuration",
            "unix_time": unix_time
        }
        status = self._check_command_response_status(request)
        return status
		
    def modem_configuration(self, host, port):
        """Set the host:port for the Cellular device to send data to.

        Returns True if the command was successful.
        """
        request = {
            "command": "modem_configuration",
            "host": host,
            "port": port
        }
        status = self._check_command_response_status(request)
        return status

    def set_acceptance_filter_bypass(self, bus, bypass):
        """Control the status of CAN acceptance filter for a bus.

        Returns True if the command was successful.
        """
        request = {
            "command": "af_bypass",
            "bus": bus,
            "bypass": bypass
        }
        return self._check_command_response_status(request)

    def set_predefined_obd2_requests(self, enabled):
        """Control if pre-defined OBD2 requests should be sent.

        Returns True if the command was successful.
        """
        request = {
            "command": "predefined_obd2",
            "enabled": enabled
        }
        return self._check_command_response_status(request)

    def _check_command_response_message(self, request):
        responses = self.complex_request(request)
        result = None
        if len(responses) > 0:
            result = responses[0].get('message')
        return result

    def version(self):
        """Request a firmware version identifier from the VI.
        """
        request = {
            "command": "version"
        }
        return self._check_command_response_message(request)

    def sd_mount_status(self):
        """Request for SD Mount status if available.
        """
        request = {
            "command": "sd_mount_status"
        }
        responses = self.complex_request(request)
        result = None
        if len(responses) > 0:
            result = responses[0].get('status')
        return result
		
    def device_id(self):
        """Request the unique device ID of the attached VI.
        """
        request = {
            "command": "device_id"
        }
        return self._check_command_response_message(request)

    def write(self, **kwargs):
        """Serialize a raw or translated write request and send it to the VI,
        following the OpenXC message format.
        """
        if 'id' in kwargs and 'data' in kwargs:
            result = self.write_raw(kwargs['id'], kwargs['data'],
                    bus=kwargs.get('bus', None),
                    frame_format=kwargs.get('frame_format', None))
        else:
            result = self.write_translated(kwargs['name'], kwargs['value'],
                    event=kwargs.get('event', None))
        return result

    def write_translated(self, name, value, event=None):
        """Send a translated write request to the VI.
        """
        data = {'name': name}
        if value is not None:
            data['value'] = self._massage_write_value(value)
        if event is not None:
            data['event'] = self._massage_write_value(event);
        message = self.streamer.serialize_for_stream(data)
        bytes_written = self.write_bytes(message)
        assert bytes_written == len(message)
        return bytes_written

    def write_raw(self, message_id, data, bus=None, frame_format=None):
        """Send a raw write request to the VI.
        """
        if not isinstance(message_id, numbers.Number):
            try:
                message_id = int(message_id, 0)
            except ValueError:
                raise ValueError("ID must be numerical")
        data = {'id': message_id, 'data': data}
        if bus is not None:
            data['bus'] = bus
        if frame_format is not None:
            data['frame_format'] = frame_format
        message = self.streamer.serialize_for_stream(data)
        bytes_written = self.write_bytes(message)
        assert bytes_written == len(message)
        return bytes_written

    def stop(self):
        pass

    def write_bytes(self, data):
        """Write the bytes in ``data`` to the controller interface."""
        raise NotImplementedError("Don't use Controller directly")

    @classmethod
    def _massage_write_value(cls, value):
        """Convert string values from command-line arguments into first-order
        Python boolean and float objects, if applicable.
        """
        if not isinstance(value, numbers.Number):
            if value == "true":
                value = True
            elif value == "false":
                value = False
            elif value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            else:
                try:
                    value = float(value)
                except ValueError:
                    pass
        return value


class ControllerError(Exception):
    pass



"""Combinations of DataSource with Controller types for duplex vehicle
interfaces."""
from .sources import UsbDataSource, SerialDataSource, NetworkDataSource, \
        BluetoothVehicleInterface
from .controllers import UsbControllerMixin, SerialControllerMixin, Controller


class UsbVehicleInterface(UsbDataSource, UsbControllerMixin):
    """This class is compatibile with an OpenXC vehicle interface vehicle interface
    attached via USB. It supports full duplex reads and writes.
    """
    pass

class SerialVehicleInterface(SerialDataSource, SerialControllerMixin):
    """This class is compatibile with an OpenXC vehicle interface vehicle interface
    connected via a virtual serial port (e.g. FTDI or Bluetooth). It has full
    read support and limited write support (no control commands are supported).
    """
    pass

class NetworkVehicleInterface(NetworkDataSource, Controller):
    """This class is compatibile with an OpenXC vehicle interface vehicle interface
    connected via the network (e.g. Ethernet or Wi-Fi). It has full
    read support and limited write support (no control commands are supported).
    """
    pass



import sys
from setuptools import setup, find_packages

from openxc.version import get_version

# This is a workaround for an odd exception that occurs when running the tests.
try:
    import multiprocessing
except ImportError:
    pass

long_description = open('README.rst').read()

install_reqs = ['pyusb==1.0.0a3', 'units >= 0.5', 'argparse', 'requests==2.4.1',
        'protobuf==2.6.1']

setup(name='openxc',
    version=get_version(),
    description='A Python library to connect to an OpenXC vehicle interface',
    long_description=long_description,
    author='Christopher Peplin',
    author_email='cpeplin@ford.com',
    license='BSD',
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development',
    ],
    url='http://github.com/openxc/openxc-python',
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={'openxc': ['generator/signals.cpp*']},
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=install_reqs,
    extras_require = {
        'serial': ["pyserial"],
        'bluetooth': ["pybluez"],
        'lxml': ["lxml"],
    },
    entry_points={
        'console_scripts': [
            'openxc-dashboard = openxc.tools.dashboard:main',
            'openxc-dump = openxc.tools.dump:main',
            'openxc-control = openxc.tools.control:main',
            'openxc-gps = openxc.tools.gps:main',
            'openxc-trace-split = openxc.tools.tracesplit:main',
            'openxc-generate-firmware-code = openxc.tools.generate_code:main',
            'openxc-diag = openxc.tools.diagnostics:main',
            'openxc-scanner = openxc.tools.scanner:main',
            'openxc-obd2scanner = openxc.tools.obd2scanner:main',
        ]
    },
)



"""Vehicle data measurement types pre-defined in OpenXC."""
import numbers
import sys
import inspect

import openxc.units as units
from .utils import Range, AgingData

try:
    unicode
except NameError:
    # Python 3
    basestring = unicode = str


def all_measurements():
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and Measurement in obj.mro() and hasattr(
                obj, 'name'):
            yield obj


class Measurement(AgingData):
    """The Measurement is the base type of all values read from an OpenXC
    vehicle interface. All values encapsulated in a Measurement have an
    associated scalar unit (e.g. meters, degrees, etc) to avoid crashing a rover
    into Mars.
    """
    name = 'generic'
    DATA_TYPE = numbers.Number
    _measurement_map = {}
    unit = units.Undefined

    def __init__(self, name, value, event=None, override_unit=False, **kwargs):
        """Construct a new Measurement with the given name and value.

        Args:
            name (str):  The Measurement's generic name in OpenXC.
            value (str, float, or bool): The Measurement's value.

        Kwargs:
           event (str, bool): An optional event for compound Measurements.
           override_unit (bool): The value will be coerced to the correct units
               if it is a plain number.

        Raises:
            UnrecognizedMeasurementError if the value is not the correct units,
            e.g. if it's a string and we're expecting a numerical value
        """
        super(Measurement, self).__init__()
        self.name = name
        if (not isinstance(value, bool) and isinstance(value, numbers.Number)
                and override_unit):
            value = self.unit(value)
        self.value = value
        self.event = event

    def __repr__(self):
        return str(self.value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        """Set the value of this measurement.

        Raises:
            AttributeError: if the new value isn't of the correct units.
        """
        if self.unit != units.Undefined and new_value.unit != self.unit:
            raise AttributeError("%s must be in %s" % (
                self.__class__, self.unit))
        self._value = new_value

    @classmethod
    def from_dict(cls, data):
        """Create a new Measurement subclass instance using the given dict.

        If Measurement.name_from_class was previously called with this data's
        associated Measurement sub-class in Python, the returned object will be
        an instance of that sub-class. If the measurement name in ``data`` is
        unrecognized, the returned object will be of the generic ``Measurement``
        type.

        Args:
            data (dict): the data for the new measurement, including at least a
                name and value.
        """

        args = []

        if 'id' in data and 'data' in data:
            measurement_class = CanMessage
            args.append("Bus %s: 0x%x" % (data.get('bus', '?'), data['id']))
            args.append(data['data'])
            # TODO grab bus
        else:
            measurement_class = cls._class_from_name(data['name'])
            if measurement_class == Measurement:
                args.append(data['name'])
            args.append(data['value'])

        return measurement_class(*args, event=data.get('event', None),
                override_unit=True)

    @classmethod
    def name_from_class(cls, measurement_class):
        """For a given measurement class, return its generic name.

        The given class is expected to have a ``name`` attribute, otherwise this
        function will raise an execption. The point of using this method instead
        of just trying to grab that attribute in the application is to cache
        measurement name to class mappings for future use.

        Returns:
            the generic OpenXC name for a measurement class.

        Raise:
            UnrecognizedMeasurementError: if the class does not have a valid
                generic name
        """
        if not getattr(cls, '_measurements_initialized', False):
            cls._measurement_map = dict((m.name, m) for m in all_measurements())
            cls._measurements_initialized = True

        try:
            name = getattr(measurement_class, 'name')
        except AttributeError:
            raise UnrecognizedMeasurementError("No 'name' attribute in %s" %
                    measurement_class)
        else:
            cls._measurement_map[name] = measurement_class
            return name

    @classmethod
    def _class_from_name(cls, measurement_name):
        return cls._measurement_map.get(measurement_name, Measurement)


class NamedMeasurement(Measurement):
    """A NamedMeasurement has a class-level ``name`` variable and thus the
    ``name`` argument is not required in its constructor.
    """
    def __init__(self, value, **kwargs):
        super(NamedMeasurement, self).__init__(self.name, value, **kwargs)


class NumericMeasurement(NamedMeasurement):
    """A NumericMeasurement must have a numeric value and thus a valid range of
    acceptable values.
    """
    valid_range = None

    def percentage_within_range(self):
        return ((self.value.num - self.valid_range.min) /
                float(self.valid_range.spread)) * 100

    def within_range(self):
        return self.valid_range.within_range(self.value.num)


class StatefulMeasurement(NamedMeasurement):
    """Must have a class-level ``states`` member that defines a set of valid
    string states for this measurement's value.
    """
    DATA_TYPE = unicode
    states = None

    def valid_state(self):
        """Determine if the current state is valid, given the class' ``state``
        member.

        Returns:
            True if the value is a valid state.
        """
        return self.states is not None and self.value in self.states


class BooleanMeasurement(NamedMeasurement):
    DATA_TYPE = bool


class EventedMeasurement(StatefulMeasurement):
    DATA_TYPE = unicode


class PercentageMeasurement(NumericMeasurement):
    valid_range = Range(0, 100)
    unit = units.Percentage


class AcceleratorPedalPosition(PercentageMeasurement):
    name = "accelerator_pedal_position"

class FuelLevel(PercentageMeasurement):
    name = "fuel_level"


class VehicleSpeed(NumericMeasurement):
    name = "vehicle_speed"
    valid_range = Range(0, 321)
    unit = units.KilometersPerHour

class EngineSpeed(NumericMeasurement):
    name = "engine_speed"
    valid_range = Range(0, 8000)
    unit = units.RotationsPerMinute

class FuelConsumed(NumericMeasurement):
    name = "fuel_consumed_since_restart"
    valid_range = Range(0, 100)
    unit = units.Litre

class Latitude(NumericMeasurement):
    name = "latitude"
    valid_range = Range(-90, 90)
    unit = units.Degree

class Longitude(NumericMeasurement):
    name = "longitude"
    valid_range = Range(-180, 180)
    unit = units.Degree

class Odometer(NumericMeasurement):
    name = "odometer"
    valid_range = Range(0, 1000000)
    unit = units.Kilometer

class SteeringWheelAngle(NumericMeasurement):
    name = "steering_wheel_angle"
    valid_range = Range(-600, 600)
    unit = units.Degree

class TorqueAtTransmission(NumericMeasurement):
    name = "torque_at_transmission"
    valid_range = Range(-800, 1500)
    unit = units.NewtonMeter

class LateralAcceleration(NumericMeasurement):
    name = "lateral_acceleration"
    valid_range = Range(-5, 5)
    unit = units.MetersPerSecondSquared

class LongitudinalAcceleration(NumericMeasurement):
    name = "longitudinal_acceleration"
    valid_range = Range(-5, 5)
    unit = units.MetersPerSecondSquared


class BrakePedalStatus(BooleanMeasurement):
    name = "brake_pedal_status"

class HeadlampStatus(BooleanMeasurement):
    name = "headlamp_status"

class HighBeamStatus(BooleanMeasurement):
    name = "high_beam_status"

class ParkingBrakeStatus(BooleanMeasurement):
    name = "parking_brake_status"

class WindshieldWiperStatus(BooleanMeasurement):
    name = "windshield_wiper_status"


class IgnitionStatus(StatefulMeasurement):
    name = "ignition_status"
    states = ['off', 'accessory', 'run', 'start']

class TransmissionGearPosition(StatefulMeasurement):
    name = "transmission_gear_position"
    states = ['first', 'second', 'third', 'fourth', 'fifth', 'sixth', 'seventh',
            'eighth', 'neutral', 'reverse', 'park']

class TurnSignalStatus(StatefulMeasurement):
    name = "turn_signal_status"


class ButtonEvent(EventedMeasurement):
    name = "button_event"
    states = ['up', 'down', 'left', 'right', 'ok']

class DoorStatus(EventedMeasurement):
    name = "door_status"
    states = ['driver', 'rear_left', 'rear_right', 'passenger']


class CanMessage(Measurement):
    name = 'can_message'


class UnrecognizedMeasurementError(Exception):
    pass





from threading import Thread
import time
import logging
import requests

from openxc.formats import JsonFormatter
from .queued import QueuedSink

LOG = logging.getLogger(__name__)


class UploaderSink(QueuedSink):
    """Uploads all incoming vehicle data to a remote web application via HTTP.
    
    TODO document service side format
    """
    UPLOAD_BATCH_SIZE = 25
    HTTP_TIMEOUT = 5000

    def __init__(self, url):
        """Args:
            url (str) - the URL to send an HTTP POST request with vehicle data
        """
        super(UploaderSink, self).__init__()
        self.recorder = self.Uploader(self.queue, url)

    class Uploader(Thread):
        def __init__(self, queue, url):
            super(UploaderSink.Uploader, self).__init__()
            self.daemon = True
            self.queue = queue
            self.records = []
            self.url = url
            self.start()

        @classmethod
        def _upload(cls, url, records):
            payload = JsonFormatter.serialize(records)
            response = requests.post(url, data=payload)

            if response.status_code != requests.codes.created:
                LOG.warn("Unable to upload %d records, received %d status "
                        "from %s", len(records), response.status_code, url)
            else:
                LOG.debug("Uploaded %d records (status %d)", len(records),
                        response.status_code)

        def run(self):
            while True:
                message, _ = self.queue.get()

                message['timestamp'] = time.time()
                self.records.append(message)
                if len(self.records) > UploaderSink.UPLOAD_BATCH_SIZE:
                    self._upload(self.url, self.records)

                self.queue.task_done()



"""Binary (Protobuf) formatting utilities."""
from __future__ import absolute_import

import binascii
import numbers
import logging

import google.protobuf.message
from google.protobuf.internal.decoder import _DecodeVarint
from google.protobuf.internal import encoder

from openxc.formats.base import VehicleMessageStreamer
from openxc import openxc_pb2

LOG = logging.getLogger(__name__)

class UnrecognizedBinaryCommandError(Exception): pass

class ProtobufStreamer(VehicleMessageStreamer):
    MAX_PROTOBUF_MESSAGE_LENGTH = 200

    def parse_next_message(self):
        message = None
        remainder = self.message_buffer
        message_data = ""

        # 1. decode a varint from the top of the stream
        # 2. using that as the length, if there's enough in the buffer, try and
        #       decode try and decode a VehicleMessage after the varint
        # 3. if it worked, great, we're oriented in the stream - continue
        # 4. if either couldn't be parsed, skip to the next byte and repeat
        while message is None and len(self.message_buffer) > 1:
            message_length, message_start = _DecodeVarint(self.message_buffer, 0)
            # sanity check to make sure we didn't parse some huge number that's
            # clearly not the length prefix
            if message_length > self.MAX_PROTOBUF_MESSAGE_LENGTH:
                self.message_buffer = self.message_buffer[1:]
                continue

            if message_start + message_length > len(self.message_buffer):
                break

            message_data = self.message_buffer[message_start:message_start +
                    message_length]
            remainder = self.message_buffer[message_start + message_length:]

            message = ProtobufFormatter.deserialize(message_data)
            if message is None:
                self.message_buffer = self.message_buffer[1:]

        self.message_buffer = remainder
        return message

    def serialize_for_stream(self, message):
        protobuf_message = ProtobufFormatter.serialize(message)
        delimiter = encoder._VarintBytes(len(protobuf_message))
        return delimiter + protobuf_message


class ProtobufFormatter(object):
    @classmethod
    def deserialize(cls, data):
        message = openxc_pb2.VehicleMessage()
        try:
            message.ParseFromString(data)
        except google.protobuf.message.DecodeError as e:
            pass
        except UnicodeDecodeError as e:
            LOG.warn("Unable to parse protobuf: %s", e)
        else:
            return cls._protobuf_to_dict(message)

    @classmethod
    def serialize(cls, data):
        return cls._dict_to_protobuf(data).SerializeToString()

    @classmethod
    def _command_string_to_protobuf(self, command_name):
        if command_name == "version":
            return openxc_pb2.ControlCommand.VERSION
        elif command_name == "device_id":
            return openxc_pb2.ControlCommand.DEVICE_ID
        elif command_name == "diagnostic_request":
            return openxc_pb2.ControlCommand.DIAGNOSTIC
        elif command_name == "passthrough":
            return openxc_pb2.ControlCommand.PASSTHROUGH
        elif command_name == "af_bypass":
            return openxc_pb2.ControlCommand.ACCEPTANCE_FILTER_BYPASS
        elif command_name == "payload_format":
            return openxc_pb2.ControlCommand.PAYLOAD_FORMAT
        elif command_name == "predefined_obd2":
            return openxc_pb2.ControlCommand.PREDEFINED_OBD2_REQUESTS
        else:
            raise UnrecognizedBinaryCommandError(command_name)

    @classmethod
    def _dict_to_protobuf(cls, data):
        message = openxc_pb2.VehicleMessage()
        if 'command' in data:
            command_name = data['command']
            message.type = openxc_pb2.VehicleMessage.CONTROL_COMMAND
            message.control_command.type = cls._command_string_to_protobuf(command_name)
            if message.control_command.type == openxc_pb2.ControlCommand.PASSTHROUGH:
                message.control_command.passthrough_mode_request.bus = data['bus']
                message.control_command.passthrough_mode_request.enabled = data['enabled']
            elif message.control_command.type == openxc_pb2.ControlCommand.ACCEPTANCE_FILTER_BYPASS:
                message.control_command.acceptance_filter_bypass_command.bus = data['bus']
                message.control_command.acceptance_filter_bypass_command.bypass = data['bypass']
            elif message.control_command.type == openxc_pb2.ControlCommand.PREDEFINED_OBD2_REQUESTS:
                message.control_command.predefined_obd2_requests_command.enabled = data['enabled']
            elif message.control_command.type == openxc_pb2.ControlCommand.PAYLOAD_FORMAT:
                if data['format'] == "json":
                    message.control_command.payload_format_command.format = openxc_pb2.PayloadFormatCommand.JSON
                elif data['format'] == "protobuf":
                    message.control_command.payload_format_command.format = openxc_pb2.PayloadFormatCommand.PROTOBUF
            elif message.control_command.type == openxc_pb2.ControlCommand.DIAGNOSTIC:
                request_command = message.control_command.diagnostic_request
                action = data['action']
                if action == "add":
                    request_command.action = openxc_pb2.DiagnosticControlCommand.ADD
                elif action == "cancel":
                    request_command.action = openxc_pb2.DiagnosticControlCommand.CANCEL
                request = request_command.request
                request_data = data['request']
                request.bus = request_data['bus']
                request.message_id = request_data['id']
                request.mode = request_data['mode']
                if 'frequency' in request_data:
                    request.frequency = request_data['frequency']
                if 'name' in request_data:
                    request.name = request_data['name']
                if 'multiple_responses' in request_data:
                    request.multiple_responses = request_data['multiple_responses']
                if 'pid' in request_data:
                    request.pid = request_data['pid']
                if 'payload' in request_data:
                    request.payload = binascii.unhexlify(request_data['payload'].split('0x')[1])
        elif 'command_response' in data:
            message.type = openxc_pb2.VehicleMessage.COMMAND_RESPONSE
            message.command_response.type = cls._command_string_to_protobuf(data['command_response'])
            if 'message' in data:
                message.command_response.message = data['message']
            message.command_response.status = data['status']
        elif 'id' in data and 'data' in data:
            message.type = openxc_pb2.VehicleMessage.CAN
            if 'bus' in data:
                message.can_message.bus = data['bus']
            if 'frame_format' in data:
                if data['frame_format'] == "standard":
                    message.can_message.frame_format = openxc_pb2.RawMessage.STANDARD
                elif data['frame_format'] == "extended":
                    message.can_message.frame_format = openxc_pb2.RawMessage.EXTENDED
            message.can_message.id = data['id']
            message.can_message.data = binascii.unhexlify(data['data'].split('0x')[1])
        elif 'id' in data and 'bus' in data and 'mode' in data:
            message.type = openxc_pb2.VehicleMessage.DIAGNOSTIC
            response = message.diagnostic_response
            response.bus = data['bus']
            response.message_id = data['id']
            response.mode = data['mode']
            if 'pid' in data:
                response.pid = data['pid']
            if 'success' in data:
                response.success = data['success']
            if 'negative_response_code' in data:
                response.negative_response_code = data['negative_response_code']
            if 'value' in data:
                response.value = data['value']
            if 'payload' in data:
                response.payload = binascii.unhexlify(data['payload'].split('0x')[1])
        elif 'name' in data and 'value' in data:
            message.type = openxc_pb2.VehicleMessage.SIMPLE
            message.simple_message.name = data['name']
            value = data['value']
            if isinstance(value, bool):
                message.simple_message.value.type = openxc_pb2.DynamicField.BOOL
                message.simple_message.value.boolean_value = value
            elif isinstance(value, str):
                message.simple_message.value.type = openxc_pb2.DynamicField.STRING
                message.simple_message.value.string_value = value
            elif isinstance(value, numbers.Number):
                message.simple_message.value.type = openxc_pb2.DynamicField.NUM
                message.simple_message.value.numeric_value = value

            if 'event' in data:
                event = data['event']
                # TODO holy repeated code, batman. this will be easier to DRY
                # when https://github.com/openxc/openxc-message-format/issues/19
                # is resolved
                if isinstance(event, bool):
                    message.simple_message.event.type = openxc_pb2.DynamicField.BOOL
                    message.simple_message.event.boolean_value = event
                elif isinstance(event, str):
                    message.simple_message.event.type = openxc_pb2.DynamicField.STRING
                    message.simple_message.event.string_value = event
                elif isinstance(event, numbers.Number):
                    message.simple_message.event.type = openxc_pb2.DynamicField.NUM
                    message.simple_message.event.numeric_value = event
        return message

    @classmethod
    def _protobuf_to_dict(cls, message):
        parsed_message = {}
        if message is not None:
            if message.type == message.CAN and message.HasField('can_message'):
                can_message = message.can_message
                if can_message.HasField('bus'):
                    parsed_message['bus'] = can_message.bus
                if can_message.HasField('id'):
                    parsed_message['id'] = can_message.id
                if can_message.HasField('data'):
                    parsed_message['data'] = "0x%s" % binascii.hexlify(can_message.data)
                if can_message.HasField('frame_format'):
                    if can_message.frame_format == openxc_pb2.RawMessage.STANDARD:
                        parsed_message['frame_format'] = "standard"
                    elif can_message.frame_format == openxc_pb2.RawMessage.EXTENDED:
                        parsed_message['frame_format'] = "extended"
            elif message.type == message.DIAGNOSTIC:
                diagnostic_message = message.diagnostic_response
                if diagnostic_message.HasField('bus'):
                    parsed_message['bus'] = diagnostic_message.bus
                if diagnostic_message.HasField('message_id'):
                    parsed_message['id'] = diagnostic_message.message_id
                if diagnostic_message.HasField('mode'):
                    parsed_message['mode'] = diagnostic_message.mode
                if diagnostic_message.HasField('pid'):
                    parsed_message['pid'] = diagnostic_message.pid
                if diagnostic_message.HasField('success'):
                    parsed_message['success'] = diagnostic_message.success
                if diagnostic_message.HasField('value'):
                    parsed_message['value'] = diagnostic_message.value
                if diagnostic_message.HasField('negative_response_code'):
                    parsed_message['negative_response_code'] = diagnostic_message.negative_response_code
                if diagnostic_message.HasField('payload'):
                    parsed_message['payload'] = "0x%s" % binascii.hexlify(diagnostic_message.payload)
            elif message.type == message.SIMPLE:
                simple_message = message.simple_message
                parsed_message['name'] = simple_message.name
                if simple_message.HasField('event'):
                    event = simple_message.event
                    if event.HasField('numeric_value'):
                        parsed_message['event'] = event.numeric_value
                    elif event.HasField('boolean_value'):
                        parsed_message['event'] = event.boolean_value
                    elif event.HasField('string_value'):
                        parsed_message['event'] = event.string_value

                if simple_message.HasField('value'):
                    value = simple_message.value
                    if value.HasField('numeric_value'):
                        parsed_message['value'] = value.numeric_value
                    elif value.HasField('boolean_value'):
                        parsed_message['value'] = value.boolean_value
                    elif value.HasField('string_value'):
                        parsed_message['value'] = value.string_value
                    else:
                        parsed_message = None
                else:
                    parsed_message = None
            elif message.type == message.CONTROL_COMMAND:
                command = message.control_command
                if command.type == openxc_pb2.ControlCommand.VERSION:
                    parsed_message['command'] = "version"
                elif command.type == openxc_pb2.ControlCommand.DEVICE_ID:
                    parsed_message['command'] = "device_id"
                elif command.type == openxc_pb2.ControlCommand.DIAGNOSTIC:
                    parsed_message['command'] = "diagnostic_request"
                    parsed_message['request'] = {}
                    action = command.diagnostic_request.action
                    if action == openxc_pb2.DiagnosticControlCommand.ADD:
                        parsed_message['action'] = "add"
                    elif action == openxc_pb2.DiagnosticControlCommand.CANCEL:
                        parsed_message['action'] = "cancel"

                    request = command.diagnostic_request.request
                    parsed_message['request']['id'] = request.message_id
                    parsed_message['request']['bus'] = request.bus
                    parsed_message['request']['mode'] = request.mode

                    if request.HasField('frequency'):
                        parsed_message['request']['frequency'] = request.frequency
                    if request.HasField('name'):
                        parsed_message['request']['name'] = request.name
                    if request.HasField('multiple_responses'):
                        parsed_message['request']['multiple_responses'] = request.multiple_responses
                    if request.HasField('pid'):
                        parsed_message['request']['pid'] = request.pid
                    if request.HasField('payload'):
                        parsed_message['request']['payload'] = "0x%s" % binascii.hexlify(request.payload)
                elif command.type == openxc_pb2.ControlCommand.PASSTHROUGH:
                    parsed_message['command'] = "passthrough"
                    parsed_message['bus'] = command.passthrough_mode_request.bus
                    parsed_message['enabled'] = command.passthrough_mode_request.enabled
                elif command.type == openxc_pb2.ControlCommand.PREDEFINED_OBD2_REQUESTS:
                    parsed_message['command'] = "predefined_obd2"
                    parsed_message['enabled'] = command.predefined_obd2_requests_command.enabled
                elif command.type == openxc_pb2.ControlCommand.ACCEPTANCE_FILTER_BYPASS:
                    parsed_message['command'] = "af_bypass"
                    parsed_message['bus'] = command.acceptance_filter_bypass_command.bus
                    parsed_message['bypass'] = command.acceptance_filter_bypass_command.bypass
                elif command.type == openxc_pb2.ControlCommand.PAYLOAD_FORMAT:
                    parsed_message['command'] = "payload_format"
                    if command.payload_format_command.format == openxc_pb2.PayloadFormatCommand.JSON:
                        parsed_message['format'] = "json"
                    elif command.payload_format_command.format == openxc_pb2.PayloadFormatCommand.PROTOBUF:
                        parsed_message['format'] = "protobuf"
            elif message.type == message.COMMAND_RESPONSE:
                response = message.command_response
                if response.type == openxc_pb2.ControlCommand.VERSION:
                    parsed_message['command_response'] = "version"
                elif response.type == openxc_pb2.ControlCommand.DEVICE_ID:
                    parsed_message['command_response'] = "device_id"
                elif response.type == openxc_pb2.ControlCommand.DIAGNOSTIC:
                    parsed_message['command_response'] = "diagnostic_request"
                elif response.type == openxc_pb2.ControlCommand.PASSTHROUGH:
                    parsed_message['command_response'] = "passthrough"
                elif response.type == openxc_pb2.ControlCommand.PAYLOAD_FORMAT:
                    parsed_message['command_response'] = "payload_format"
                elif response.type == openxc_pb2.ControlCommand.ACCEPTANCE_FILTER_BYPASS:
                    parsed_message['command_response'] = "af_bypass"
                elif response.type == openxc_pb2.ControlCommand.PREDEFINED_OBD2_REQUESTS:
                    parsed_message['command_response'] = "predefined_obd2"
                else:
                    raise UnrecognizedBinaryCommandError(response.type)

                parsed_message['status'] = response.status
                if response.HasField('message'):
                    parsed_message['message'] = response.message
            else:
                parsed_message = None
        return parsed_message



"""A network socket data source."""
from __future__ import absolute_import

import logging
import socket

from .base import DataSourceError
from .socket import SocketDataSource

LOG = logging.getLogger(__name__)


class NetworkDataSource(SocketDataSource):
    """A data source reading from a network socket, as implemented
    in the openxc-vehicle-simulator .
    """
    DEFAULT_PORT = 50001

    def __init__(self, host=None, port=None, **kwargs):
        """Initialize a connection to the network socket.

        Kwargs:
            host - optionally override the default network host (default is local machine)
            port - optionally override the default network port (default is 50001)
            log_mode - optionally record or print logs from the network source

        Raises:
            DataSourceError if the socket connection cannot be opened.
        """
        super(NetworkDataSource, self).__init__(**kwargs)
        self.host = host or socket.gethostbyname(socket.gethostname())
        self.port = port or self.DEFAULT_PORT
        self.port = int(self.port)

        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((self.host, self.port))
        except (OSError, socket.error) as e:
            raise DataSourceError("Unable to open socket connection at  "
                    "%s:%s: %s" % (self.host,self.port, e))
        else:
            LOG.debug("Opened socket connection at %s:%s", self.host, self.port)



""" This module contains the methods for the ``openxc-dashboard`` command line
program.

`main` is executed when ``openxc-dashboard`` is run, and all other callables in
this module are internal only.
"""
from __future__ import absolute_import
from __future__ import division

import argparse
import curses
import math
from datetime import datetime
from threading import Lock

from .common import device_options, configure_logging, select_device
from openxc.vehicle import Vehicle
from openxc.measurements import EventedMeasurement, Measurement

try:
    unicode
except NameError:
    # Python 3
    basestring = unicode = str


# timedelta.total_seconds() is only in 2.7, so we backport it here for 2.6
def total_seconds(delta):
    return (delta.microseconds + (delta.seconds
        + delta.days * 24 * 3600) * 10**6) / 10**6


# Thanks, SO: http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
def sizeof_fmt(num):
    for unit in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, unit)
        num /= 1024.0


class DataPoint(object):
    AVERAGE_FREQUENCY_ALPHA = 0.1

    def __init__(self, measurement_type):
        self.event = ''
        self.current_data = None
        self.events = {}
        self.messages_received = 0
        self.measurement_type = measurement_type
        self.min = None
        self.max = None
        self.last_update_time = None
        self.average_time_since_update = None

    def update(self, measurement):
        self.messages_received += 1
        self.current_data = measurement

        if self.last_update_time is not None:
            time_since_update = total_seconds(datetime.now() - self.last_update_time)
            if self.average_time_since_update is None:
                self.average_time_since_update = time_since_update
            else:
                self.average_time_since_update = ((self.AVERAGE_FREQUENCY_ALPHA *
                        time_since_update) + (1 - self.AVERAGE_FREQUENCY_ALPHA) *
                        self.average_time_since_update)
        self.last_update_time = datetime.now()

        if getattr(self.current_data.value, 'unit', None) == self.current_data.unit:
            if self.min is None or self.current_data.value < self.min:
                self.min = self.current_data.value
            elif self.max is None or self.current_data.value > self.max:
                self.max = self.current_data.value

        if isinstance(measurement, EventedMeasurement):
            if measurement.valid_state():
                self.events[measurement.value] = measurement.event

    def percentage(self):
        # TODO man, this is getting really ugly to handle all of the different
        # types
        percent = None
        if hasattr(self.measurement_type, 'valid_range'):
            percent = self.current_data.percentage_within_range()
        elif (getattr(self, 'min', None) is not None and
                getattr(self, 'max', None) is not None) and self.min != self.max:
            percent = (((self.current_data.value - self.min) / float(self.max -
                    self.min)) * 100).num
        return percent

    def print_to_window(self, window, row, started_time):
        width = window.getmaxyx()[1]
        if self.current_data is not None:
            if len(self.events) == 0:
                value = str(self.current_data)
            else:
                result = ""
                for item, value in enumerate(self.measurement_type.states):
                    # TODO missing keys here?
                    result += "%s: %s " % (value, self.events.get(value, "?"))
                value = result

            value_color = curses.color_pair(2)
            window.addstr(row, 0, "{:>22} %s".format(value), value_color)
        window.addstr(row, 23, self.current_data.name)

        if width > 60:
            message_count_color = curses.color_pair(3)
            window.addstr(row, 50, "Msgs: " + str(self.messages_received),
                    message_count_color)

        if width > 75 and self.average_time_since_update > 0:
            window.addstr(row, 61, "Freq. (Hz): %d" %
                    math.ceil((1 / self.average_time_since_update)))


class Dashboard(object):
    def __init__(self, window, vehicle):
        self.window = window
        self.elements = {}
        self.scroll_position = 0
        self.screen_lock = Lock()
        vehicle.listen(Measurement, self.receive)

        self.started_time = datetime.now()
        self.messages_received = 0

        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_RED, -1)
        curses.init_pair(2, curses.COLOR_GREEN, -1)
        curses.init_pair(3, curses.COLOR_YELLOW, -1)

    def receive(self, measurement, **kwargs):
        if self.messages_received == 0:
            self.started_time = datetime.now()
        self.messages_received += 1

        if measurement.name not in self.elements:
            self.elements[measurement.name] = DataPoint(measurement.__class__)
        self.elements[measurement.name].update(measurement)
        self._redraw()

    def _redraw(self):
        self.screen_lock.acquire()
        self.window.erase()
        max_rows = self.window.getmaxyx()[0] - 4
        for row, element in enumerate(sorted(self.elements.values(),
                key=lambda elt: elt.current_data.name)[self.scroll_position:]):
            if row > max_rows:
                break
            element.print_to_window(self.window, row, self.started_time)
        self.window.addstr(max_rows + 1, 0,
                "Message count: %d (%d corrupted)" % (self.messages_received,
                    self.source.corrupted_messages), curses.A_REVERSE)
        self.window.addstr(max_rows + 2, 0,
                "Total received: %s" %
                sizeof_fmt(self.source.bytes_received),
                curses.A_REVERSE)
        self.window.addstr(max_rows + 3, 0, "Data Rate: %s" %
            sizeof_fmt(self.source.bytes_received /
                (total_seconds(datetime.now() - self.started_time)
                    + 0.1)),
             curses.A_REVERSE)
        self.window.refresh()
        self.screen_lock.release()

    def scroll_down(self, lines):
        self.screen_lock.acquire()
        self.scroll_position = min(self.window.getmaxyx()[1],
                self.scroll_position + lines)
        self.screen_lock.release()

    def scroll_up(self, lines):
        self.screen_lock.acquire()
        self.scroll_position = max(0, self.scroll_position - lines)
        self.screen_lock.release()


def run_dashboard(window, source_class, source_kwargs):
    vehicle = Vehicle()
    dashboard = Dashboard(window, vehicle)
    dashboard.source = source_class(**source_kwargs)
    vehicle.add_source(dashboard.source)

    window.scrollok(True)
    while True:
        c = window.getch()
        if c == curses.KEY_DOWN:
            dashboard.scroll_down(1)
        elif c == curses.KEY_UP:
            dashboard.scroll_up(1)
        elif c == curses.KEY_NPAGE:
            dashboard.scroll_down(25)
        elif c == curses.KEY_PPAGE:
            dashboard.scroll_up(25)



def parse_options():
    parser = argparse.ArgumentParser(
            description="View a real-time dashboard of all OpenXC measurements",
            parents=[device_options()])
    arguments = parser.parse_args()
    return arguments


def main():
    configure_logging()
    arguments = parse_options()
    source_class, source_kwargs = select_device(arguments)
    curses.wrapper(run_dashboard, source_class, source_kwargs)



"""This module is contains the Vehicle class, which is the main entry point for
using the Python library to access vehicle data programatically. Most users will
want to interact with an instance of Vehicle, and won't need to deal with other
parts of the library directly (besides measurement types).
"""
from .measurements import Measurement
from .sinks import MeasurementNotifierSink

class Vehicle(object):
    """The Vehicle class is the main entry point for the OpenXC Python library.
    A Vehicle represents a connection to at least one vehicle data source and
    zero or 1 vehicle controllers, which can accept commands to send back to the
    vehicle. A Vehicle instance can have more than one data source (e.g. if the
    computer using this library has a secondary GPS data source).

    Most applications will either request synchronous vehicle data measurements
    using the ``get`` method or or with a callback function passed to
    ``listen``.

    More advanced applications that want access to all raw vehicle data may want
    to register a ``DataSink`` with a Vehicle.
    """

    def __init__(self, interface=None):
        """Construct a new Vehicle instance, optionally providing an vehicle
        interface from ``openxc.interface`` to user for I/O.
        """
        self.sources = set()
        self.sinks = set()
        self.measurements = {}

        if interface is not None:
            self.add_source(interface)
            self.controller = interface

        self.notifier = MeasurementNotifierSink()
        self.sinks.add(self.notifier)

    def get(self, measurement_class):
        """Return the latest measurement for the given class or None if nothing
        has been received from the vehicle.
        """
        name = Measurement.name_from_class(measurement_class)
        return self._construct_measurement(name)

    def listen(self, measurement_class, callback):
        """Register the callback function to be called whenever a new
        measurement of the given class is received from the vehicle data
        sources.

        If the callback is already registered for measurements of the given
        type, this method will have no effect.
        """
        self.notifier.register(measurement_class, callback)

    def unlisten(self, measurement_class, callback):
        """Stop notifying the given callback of new values of the measurement
        type.

        If the callback was not previously registered as a listener, this method
        will have no effect.
        """
        self.notifier.unregister(measurement_class, callback)

    def add_source(self, source):
        """Add a vehicle data source to the instance.

        The Vehicle instance will be set as the callback of the source, and the
        source will be started if it is startable. (i.e. it has a ``start()``
        method).
        """
        if source is not None:
            self.sources.add(source)
            source.callback = self._receive
            if hasattr(source, 'start'):
                source.start()

    def add_sink(self, sink):
        """Add a vehicle data sink to the instance. ``sink`` should be a
        sub-class of ``DataSink`` or at least have a ``receive(message,
        **kwargs)`` method.

        The sink will be started if it is startable. (i.e. it has a ``start()``
        method).
        """
        if sink is not None:
            self.sinks.add(sink)
            if hasattr(sink, 'start'):
                sink.start()

    def _receive(self, message, **kwargs):
        name = message.get('name', 'can_message')
        self.measurements[name] = message

        for sink in self.sinks:
            sink.receive(message, **kwargs)

    def _construct_measurement(self, measurement_id):
        raw_measurement = self.measurements.get(measurement_id, None)
        if raw_measurement is not None:
            return Measurement.from_dict(raw_measurement)
